package anywheresoftware.b4a.objects;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.util.List;
import java.util.Set;

import org.json.JSONArray;

import android.content.Intent;
import anywheresoftware.b4a.AbsObjectWrapper;
import anywheresoftware.b4a.BA;
import anywheresoftware.b4a.BA.DependsOn;
import anywheresoftware.b4a.BA.Events;
import anywheresoftware.b4a.BA.Permissions;
import anywheresoftware.b4a.BA.ShortName;
import anywheresoftware.b4a.BA.Version;
import anywheresoftware.b4a.keywords.DateTime;

import com.parse.CountCallback;
import com.parse.DeleteCallback;
import com.parse.FindCallback;
import com.parse.GetCallback;
import com.parse.GetDataCallback;
import com.parse.Parse;
import com.parse.ParseAnalytics;
import com.parse.ParseException;
import com.parse.ParseFile;
import com.parse.ParseGeoPoint;
import com.parse.ParseInstallation;
import com.parse.ParseObject;
import com.parse.ParsePush;
import com.parse.ParseQuery;
import com.parse.ParseRelation;
import com.parse.PushService;
import com.parse.RefreshCallback;
import com.parse.SaveCallback;
import com.parse.SendCallback;

/**
 *The ParseObject is a local representation of data that can be saved and retrieved from the Parse cloud.
 *The basic workflow for creating new data is to construct a new ParseObject, use Put() to fill it with data, and then use Save() to persist to the database.
 *The basic workflow for accessing existing data is to use a ParseQuery to specify which existing data to retrieve. 
 */
@Events(values={
        "DoneSave(Success As Boolean, TaskID As Int)",
        "DoneDelete(Success As Boolean, TaskID As Int)",
        "DoneRefresh(Success As Boolean, TaskID As Int)"
        })
@DependsOn(values={"ParseNative"})
@Version(1.1f)
@ShortName("ParseObject")
public class ParseObjectWrapper extends AbsObjectWrapper<ParseObject>
{

	
	private static final double version = 1.1;
	// 1.0	Erel's original beta
	// 1.1	added to ParseObject: Add, CreateWithoutData, IsDataAvailable, GetKeys, GetJSONArray 
	//		added to ParseObject: GetObject, GetObjectType, getQuery, Increment
	//
	//		added to Parse: EnableNotification, TrackOPening, Subscribe, Unsubscribe, Subscriptions
	//
	//		added to ParseQuery: WhereNear, WhereWithinKilometers, WhereWithinMiles
	//
	//		added ParseFile object
	//		added ParseGeoPoint object
	//		added ParseRelation object


	/**
	 *Returns the version of the library.
	 */
	public double getVersion()
	{
		return version;
	}
	
	/**
	 *The Parse platform provides a complete backend solution for your mobile application. 
	 *On Parse, you create an App for each of your mobile applications.
	 *Each App has its own application id and client key.
	 *Your account on Parse can accommodate multiple Apps.
	 *See the documentation on the Parse.com web site for more information.
	 *<link>https://parse.com/docs/index|https://parse.com/docs/index</link>
	 *
	 *This library supports ParseObject, ParseQquery, ParseFile, ParseGeoPoint and Push Notifications.
	 *
	 *All the operations which need to access the Parse online service are done asynchronously
	 *and raise events when complete whose parameters include a Success indication.
	 *If an error occurs Success will be False and a ParseException with a message describing the error
	 *will be placed in LastException. If necessary this should be processed in the event code
	 *as there is a possibility that a further error could occur in another asynchronous call that
	 *would overwrite the existing LastException. 
	 *
	 *Note that these asynchronous operations do not necessarily complete in the order in which they are invoked.
	 *To deal with this possibility a TaskID parameter is provided in each call that can be used, if required,
	 *to identify which of several possible method calls has just completed. This may not always be needed.
	 *
	 *
	 *Copyright 2013 Anywhere Software
	 *
	 *Licensed under the Apache License, Version 2.0 (the "License");
	 *you may not use this file except in compliance with the License.
	 *You may obtain a copy of the License at
	 *
	 *		http://www.apache.org/licenses/LICENSE-2.0
	 *
	 *Unless required by applicable law or agreed to in writing, software
	 *distributed under the License is distributed on an "AS IS" BASIS,
	 *WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
	 *See the License for the specific language governing permissions and
	 *limitations under the License.
	 */
	public void LIBRARY_DOC()
	{
	}
	
	
	/**
     *Initializes the ParseObject and sets the class name into which the Key Value pairs will be saved.
     */
    public void Initialize(String ClassName) {
        setObject(new ParseObject(ClassName));
    }    
    
    /**
     *Adds the provided Value to the array of values for Key.
     *The first use of Add for a Key defines the Value of that Key as an array
     *The array of values for a Key may be accessed by GetJSONArray.
     */
    public void Add(String Key, Object Value) {
        getObject().add(Key, Value);
    }
        
    /**
     *Returns True if the provided Key is present in the ParseObject.
     */
    public boolean ContainsKey(String Key) {
        return getObject().containsKey(Key);
    }
    
    /**
     *Places the provided Key Value pair in the present ParseObject.
     */
    public void Put(String Key, Object Value) {
        getObject().put(Key, Value);
    }
        
    /**
     *Creates a reference to an existing ParseObject for use in creating associations between ParseObjects.
     *Calling IsDataAvailable on this object will return false until Refresh has been called.
     *No network request will be made. 
     */
    public ParseObject CreateWithoutData(String ClassName, String ObjectID)
    {
    	return ParseObject.createWithoutData(ClassName, ObjectID);
    }
    
    
    /**
     *Returns a String array containing the names of the Keys in the ParseObject.
     *Only the Keys with defined data values are returned.
     */
    public String[] GetKeys() {
        Set<String> keys = getObject().keySet();
        String[] akeys = new String[0];
        return keys.toArray(akeys);        
    }

    /**
     *Returns the value of a Key as a JSON array or an empty string if the value is undefined.
     *Note that Key is case-sensitive.
     */
    public String GetJSONArray(String Key) {
        JSONArray arr  = getObject().getJSONArray(Key);
        if (arr == null)
        	return "";
        return arr.toString();        
    }  
    
    /**
     *Returns the Value of Key as a String.  Note that Key is case-sensitive.
     */
    public String GetString(String Key) {
        return getObject().getString(Key);
    }
    
    /**
     *Returns the Value of Key as a Double. Note that Key is case-sensitive.
     */
    public double GetDouble(String Key) {
        return getObject().getDouble(Key);
    }
    
    /**
     *Returns the Value of Key as a Long.
     */
    public long GetLong(String Key) {
        return getObject().getLong(Key);
    }
    
   /**
    *Returns the Value of Key as a Boolean. Note that Key is case-sensitive.
    */
   public boolean GetBoolean(String Key) {
        return getObject().getBoolean(Key);
    }
    
   /**
    *Returns the Value of Key as an Int. Note that Key is case-sensitive.
    */
    public int GetInt(String Key) {
        return getObject().getInt(Key);
    }
    
    /**
     *Returns the time when the data was first created on the Parse server.
     *This reports time as the server sees it, so that if you create a ParseObject,
     *then wait a while, and then call save(), the creation time will be the time of
     *the first save() call rather than the time the object was created locally.
     */
    public long getCreatedAt() {
        return getObject().getCreatedAt().getTime();
    }
  
    /**
     *Returns True if the ParseObject is new or has been fetched or refreshed. False otherwise.
     */
    public boolean getIsDataAvailable()
    {
    	return getObject().isDataAvailable();
    }
    
    /**
     *Returns the time when the data was last updated on the Parse server.
     *This reports time as the server sees it, so that if you make changes to a ParseObject,
     *then wait a while, and then call save(), the updated time will be the time of the save()
     *call rather than the time the object was changed locally.
     */
    public long getUpdatedAt() {
        return getObject().getUpdatedAt().getTime();
    }
    
    
    /**
     *Returns the Value object for the specified Key. Note that Key is case-sensitive.
     *In most cases it is more convenient to use a helper function such as GetString or GetInt.
     *This is mostly for use with Byte arrays, ParseFile, ParseGeoPoint and ParseObject objects. 
     */
    public Object GetObject(String Key)
    {
    	return getObject().get(Key);
    }
    
    /**
     *Returns the type name of the Value that is returned by GetObject for the specified Key.
     *Note that Key is case-sensitive. This is provided for diagnostic purposes.
     */
    public String GetObjectType(String Key)
    {
    	Object o =  getObject().get(Key);
    	if ( o == null)
    		return "null";
    	return o.getClass().getName();
    }  
    
    
    /**
     *Returns the unique ObjectId for the ParseObject.
     *An object id is assigned as soon as an object is saved to the server.
     *The combination of a Class name and an ObjectId uniquely identifies an object in your application. 
     */
    public String getObjectId() {
        return getObject().getObjectId();
    }
    
    /**
     *Access or create a Relation value for a key.
     *Returns the ParseRelation object if the relation already exists, or can be created, for the Key.
     */
    public ParseRelationWrapper GetRelation(String Key)
    {
    	ParseRelationWrapper prw = new ParseRelationWrapper();
    	prw.setObject(getObject().getRelation(Key)); 
    	return prw;    	
    }
    
    
    /**
     *Increment (or decrement) the value of the given Key by the amount Number.
     */
    public void Increment(String Key, Double Number)
    {
    	getObject().increment(Key, Number);    	
    }
    
    /**
     *Saves the object on the server.
     *The DoneSave event will be raised when saving completes.
     *If an error occurs Success will be False and a ParseException with a message describing the error will be placed in LastException.
     *If you want to ignore the event then pass an empty String for EventName.
     */
    public void Save(final BA ba, final String EventName, final int TaskID) {
        if (EventName.length() == 0) {
            getObject().saveInBackground();
        }
        else {
            getObject().saveInBackground(new SaveCallback() {
    
                @Override
                public void done(ParseException e) {
                    setLastException(ba, e);
                    ba.raiseEventFromDifferentThread(getObject(), null, 0, EventName.toLowerCase(BA.cul) + "_donesave", 
                        true,
                        new Object[] {e == null, TaskID});
                }
                
            });
        }
    }
    
    /**
     *Deletes the object from the server.
     *The DoneDelete event will be raised when saving completes.
     *If an error occurs Success will be False and a ParseException with a message describing the error will be placed in LastException.
     *If you want to ignore the event then pass an empty String for EventName.
     */
    public void Delete(final BA ba, final String EventName, final int TaskID) {
        if (EventName.length() == 0) {
            getObject().deleteInBackground();
        }
        else {
            getObject().deleteInBackground(new DeleteCallback() {

                @Override
                public void done(ParseException e) {
                    setLastException(ba, e);
                    ba.raiseEventFromDifferentThread(getObject(), null, 0, EventName.toLowerCase(BA.cul) + "_donedelete", 
                        true,
                        new Object[] {e == null, TaskID});
                }
                
            });
        }
    }
    
    /**
     *Refreshes the object by reading it from the server.
     *The DoneRefresh event will be raised when the operation completes.
     *If an error occurs Success will be False and a ParseException with a message describing the error will be placed in LastException.
    */
    public void Refresh(final BA ba, final String EventName, final int TaskID) {
        getObject().refreshInBackground(new RefreshCallback() {

            @Override
            public void done(ParseObject object, ParseException e) {
                setLastException(ba, e);
                ba.raiseEventFromDifferentThread(getObject(), null, 0, EventName.toLowerCase(BA.cul) + "_donerefresh", 
                    true,
                    new Object[] {e == null, TaskID});
            }
            
        });
    }
    
    static void setLastException(BA ba, Exception e) {
        if (e != null) {
            BA b = ba;
            if (ba.processBA != null)
                b = ba.processBA;
            b.setLastException(e);
            e.printStackTrace();
        }
    }
    
    
    
	//###########################################################################
	// Parse
	//###########################################################################
	    
    
    /**
     *This object has three methods.
     *Initialize, which is required to Initialize the Parse library.
     *TrackOpening allows openings of your applications to be recorded.
     *EnableNotifications lets your application receive push notifications.
     */
    @ShortName("Parse")
    @Permissions(values = {"android.permission.INTERNET",
    		"android.permission.ACCESS_NETWORK_STATE",
    		"android.permission.RECEIVE_BOOT_COMPLETED",
    		"android.permission.VIBRATE"  		
    })
    public static class ParseWrapper {
    	
    	/**
    	 *Returns the version of the library.
    	 */
    	public double getVersion()
    	{
    		return version;
    	}
    	
    	/**
         * Initializes the Parse library.
         */
        public void Initialize(String ApplicationId, String ClientId) {
            Parse.initialize(BA.applicationContext, ApplicationId, ClientId);
        }
        
        /**
         *Enables tracking of launches of this application.
         *
         *The following is a technical comment included as a point of record.
         *trackAppOpened does not work as documented in the Parse API.
         *It states that a null argument to ParseAnalytics.trackAppOpened will work, but the code checks for null
         *and does not send the tracking data if the intent is null or if the intent extras are null.
         *This method creates a new intent with putExtra("Data", "B4a_App") and sends that which seems to work OK.
         */
        public void TrackOpening() {
        	Intent i = new Intent();
        	i.putExtra("Data", "B4a_App");
        	ParseAnalytics.trackAppOpened(i);
        }
        
        
        /**
         *Enable push notifications and cause the specified Activity to be opened when one is received.
         *The push data may be obtained in the Activity from the StartingIntent ExtrasToString method.
         *To enable push you also need to add the following to the Manifest Editor in the project.
         *
         *The Activity named will be started for a push sent to everyone registered for push with the Parse application.
    	 *The default application icon will be used for the notification.
         *<code>
         *AddApplicationText(<service android:name="com.parse.PushService" />
         *<receiver android:name="com.parse.ParseBroadcastReceiver">
         *   <intent-filter>
         *      <action android:name="android.intent.action.BOOT_COMPLETED" />
         *      <action android:name="android.intent.action.USER_PRESENT" />
         *   </intent-filter>
         *</receiver>)</code>
         *Object Browser does not show the actual XML text within AddApplicationText above.
         *The XML is available for cut and paste in the first post of the Parse library forum thread 
         */
        @SuppressWarnings({ "unchecked", "rawtypes" })
		public void EnableNotifications(BA ba, Class ActivityName)
        {
        	PushService.setDefaultPushCallback(ba.context, ActivityName);
        	ParseInstallation.getCurrentInstallation().saveInBackground(); 
        }         
        
        /**
         *Disables receiving pushes sent to everyone
         *If no channels are subscribed to it should stop the PushService.
         */
        public void DisableNotifications(BA ba)
        {
        	PushService.setDefaultPushCallback(ba.context, null);
        }        
        
    	/**
    	 *Call this function when the user should be subscribed to a new push channel.
    	 *The default application icon will be used for the notification.
         *The Activity named will be started for a push received for the specified channel.
    	 */
    	@SuppressWarnings({ "unchecked", "rawtypes" })
		public void Subscribe(BA ba, String Channel, Class ActivityName)
    	{
    		PushService.subscribe(ba.context, Channel, ActivityName);
    	}
    	
    	/**
    	 *Call this function when the user should be subscribed to a new push channel.
    	 *The specified icon file from the application resources will be used for the notification.
         *The Activity named will be started for a push received for the specified channel.
    	 */
    	@SuppressWarnings({ "unchecked", "rawtypes" })
		public void Subscribe2(BA ba, String Channel, Class ActivityName, String Icon)
    	{
    		int icon = BA.applicationContext.getResources().getIdentifier(Icon, "drawable", BA.packageName);
    		PushService.subscribe(ba.context, Channel, ActivityName, icon);
    	}
    	
    	/**
    	 *Call this function when the user should be unsubscribed from a push channel.
    	 */
		public void Unsubscribe(BA ba, String Channel)
    	{
    		PushService.unsubscribe(ba.context, Channel);
    	}
    	
    	/**
    	 *Accesses the current set of channels for which the current installation is subscribed.
    	 */
    	public String[] Subscriptions(BA ba)
    	{
    		Set<String> subset = PushService.getSubscriptions(ba.context);
    		return subset.toArray(new String[0]);    		
    	}
    	
    	
    	
    	/**
    	 *This is included for diagnostic purposes only. You should not use this unless requested to.
    	 *
    	 *Sets the level of logging to display, where each level includes all those below it.
    	 *Ensure this is set to Parse.LOG_LEVEL_ERROR or Parse.LOG_LEVEL_NONE before deploying your app
    	 *to ensure no sensitive information is logged. The default level is Parse.LOG_LEVEL_ERROR.
    	 *
    	 *The levels are:
    	 *	LOG_LEVEL_VERBOSE = 2
    	 *	LOG_LEVEL_DEBUG = 3
    	 *	LOG_LEVEL_INFO = 4
    	 *	LOG_LEVEL_WARNING = 5
    	 *	LOG_LEVEL_ERROR = 6
    	 *	LOG_LEVEL_NONE = 2147483647;
    	 */
    	public void SetLoggingLevel(int LogLevel)
    	{
    		Parse.setLogLevel(LogLevel);
    	}
        
    }
    

    
    //###########################################################################
	// ParseQuery
	//###########################################################################
    
    @ShortName("ParseQuery")
    @Events(values={
            "DoneGet(Success As Boolean, PO As ParseObject, TaskId As Int)",
            "DoneFind(Success As Boolean, ListOfPO As List, TaskId As Int)",
            "DoneCount(Success As Boolean, Count As Int, TaskId As Int)"})
    public static class ParseQueryWrapper extends AbsObjectWrapper<ParseQuery> {
    	
    	/**
    	 *Returns the version of the library.
    	 */
    	public double getVersion()
    	{
    		return version;
    	}   	
    	
    	
        /**
         *Initializes the object and sets the Class name.
         */
        public void Initialize(String ClassName) {
            setObject(new ParseQuery(ClassName));
        }
        
        /**
         *Gets an object based on its ObjectId.
         *The DoneGet event will be raised when the operation completes.
         *If an error occurs Success will be False and a ParseException with a message describing the error will be placed in LastException.
         */
        public void GetById(final BA ba, final String EventName, final String ObjectId, final int TaskId) {
            getObject().getInBackground(ObjectId, new GetCallback() {
                @Override
                public void done(ParseObject object, ParseException e) {
                    setLastException(ba, e);
                    ba.raiseEventFromDifferentThread(getObject(), null, 0, EventName.toLowerCase(BA.cul) + "_doneget",
                            true,
                            new Object[] {e == null, object, TaskId});
                }
                
            });
        }
        
        /**
         *Finds all objects matching the query.
         *The DoneFind event will be raised when the operation completes.
         *If an error occurs Success will be False and a ParseException with a message describing the error will be placed in LastException.
         */
        public void Find(final BA ba, final String EventName, final int TaskId) {
            getObject().findInBackground( new FindCallback() {
                @Override
                public void done(List<ParseObject> objects, ParseException e) {
                    setLastException(ba, e);
                    anywheresoftware.b4a.objects.collections.List l = new anywheresoftware.b4a.objects.collections.List();
                    if (objects != null) {
                        l.Initialize();
                        l.getObject().addAll(objects);
                    }
                    ba.raiseEventFromDifferentThread(getObject(), null, 0, EventName.toLowerCase(BA.cul) + "_donefind", 
                            true,
                            new Object[] {e == null, l, TaskId});
                }
            });
        }
        
        /**
         *Counts the number of objects matching the query.
         *The DoneCount event will be raised when the operation completes.
         *If an error occurs Success will be False and a ParseException with a message describing the error will be placed in LastException.
         */
        public void Count(final BA ba, final String EventName, final int TaskId) {
            getObject().countInBackground(new CountCallback() {
                @Override
                public void done(int count, ParseException e) {
                    setLastException(ba, e);
                    ba.raiseEventFromDifferentThread(getObject(), null, 0, EventName.toLowerCase(BA.cul) + "_donecount", 
                            true,
                            new Object[] {e == null, count, TaskId});
                }
            });
        }
        
        /**
         *Sets the ordering of the query results as Ascending or Descending.
         */
        public void OrderBy (String Key, boolean Ascending) {
            if (Ascending)
                getObject().orderByAscending(Key);
            else
                getObject().orderByDescending(Key);
        }
        
        /**
         *Adds a WhereEqualTo condition to the query.
         */
        public ParseQueryWrapper WhereEqualTo(String Key, Object Value) {
            getObject().whereEqualTo(Key, Value);
            return this;
        }
        
        /**
         *Adds a WhereGreaterThan condition to the query.
         */
        public ParseQueryWrapper WhereGreaterThan(String Key, Object Value) {
            getObject().whereGreaterThan(Key, Value);
            return this;
        }
        
        /**
         *Adds a WhereLessThan condition to the query.
         */
        public ParseQueryWrapper WhereLessThan(String Key, Object Value) {
            getObject().whereLessThan(Key, Value);
            return this;
        }
        
        /**
         *Adds a WhereNotEqualTo condition to the query.
         */
        public ParseQueryWrapper WhereNotEqualTo(String Key, Object Value) {
            getObject().whereNotEqualTo(Key, Value);
            return this;
        }
        
        /**
         *Adds a WhereNear condition to the query.
         */
        public ParseQueryWrapper WhereNear(String Key, ParseGeoPoint GeoPoint) {
            getObject().whereNear(Key, GeoPoint);
            return this;
        }
        
        /**
         *Adds a WhereWithinKilometers condition to the query.
         */
        public ParseQueryWrapper WhereWithinKilometers(String Key, ParseGeoPoint GeoPoint, double Distance) {
            getObject().whereWithinKilometers(Key, GeoPoint, Distance);
            return this;
        }
        
        /**
         *Adds a WhereWithinMiles condition to the query.
         */
        public ParseQueryWrapper WhereWithinMiles(String Key, ParseGeoPoint GeoPoint, double Distance) {
            getObject().whereWithinMiles(Key, GeoPoint, Distance);
            return this;
        }      
        /**
         *Add a constraint to the query that requires a particular key's coordinates to be contained within a given rectangular geographic bounding box, 
         *southwest as ParseGeoPoint:
         *The lower-left inclusive corner of the box.
         *northeast as ParseGeoPoint:
         *The upper-right inclusive corner of the box.
         */
        public ParseQueryWrapper whereWithinGeoBox(String Key, ParseGeoPoint southwest, ParseGeoPoint northeast) {
            getObject().whereWithinGeoBox(Key, southwest, northeast);
            return this;
        }      
        /**
         *Limits the number of results. Pass -1 to retrieve all items.
         */
        public void setLimit(int v) {
            getObject().setLimit(v);
        }
        /**
         *You can skip the first results with setSkip. This can be useful for pagination.
         */
        public void setSkip(int v) {
          getObject().setSkip(v);
      }
        
    }

    
    
	//###########################################################################
	// ParseFile
	//###########################################################################
	    
    
    /**
     *ParseFile is a local representation of a file that is saved to the Parse cloud. 
     *The workflow is to construct a ParseFile with data and a filename.
     *Then save it and set it as a field on a ParseObject.
     *You can also use it to store any binary data (up to 10 megabytes). 
     */
    @Events(values={
            "DoneGet(Success As Boolean, Data() As Byte, TaskID As Int)",
            "DoneSave(Success As Boolean, TaskID as Int)"
            })
    @ShortName("ParseFile")
    @Permissions(values = {"android.permission.INTERNET"})
    public static class ParseFileWrapper extends AbsObjectWrapper<ParseFile> {
    	
    	/**
    	 *Returns the version of the library.
    	 */
    	public double getVersion()
    	{
    		return version;
    	}
    	
    	/**
         *Initializes the ParseFile with a name and the file data as a Byte array.
         *Giving a SaveName with a proper file extension (e.g. ".png") is best because it allows
         *Parse to deduce the content type of the file and set appropriate HTTP headers when it is fetched. 
         */
        public void Initialize(String SaveName, byte[] Data) {
            setObject(new ParseFile(SaveName, Data));
        }
        
    	/**
         *Initializes the ParseFile from a filename and path.
         *Giving a SaveName with a proper file extension (e.g. ".png") is best because it allows
         *Parse to deduce the content type of the file and set appropriate HTTP headers when it is fetched. 
         */
        public void Initialize2(String Path, String Name, String SaveName) throws Exception {
        	File file = new File(Path, Name);
        	byte[] bytes = new byte[(int) file.length()];
            InputStream ins = new FileInputStream(file);       	
            int offset = 0;
            int numRead = 0;
            while (offset < bytes.length
                   && (numRead=ins.read(bytes, offset, bytes.length-offset)) >= 0) {
                offset += numRead;
            } 
            ins.close();
        	setObject(new ParseFile(SaveName, bytes));
        }
        
      
        /**
         *Returns the file SaveName. Before Save is called, this is just the SaveName given by Initialize.
         *After Save is called, that name gets prefixed with a unique identifier.
         */
        public String getName()
        {
        	return getObject().getName();
        }        
        
        /**
         *Returns whether the file still needs to be saved. 
         */
        public boolean getIsDirty()
        {
        	return getObject().isDirty();
        }
        
        /**
         *Returns whether the file has available data. 
         */
        public boolean getIsDataAvailable()
        {
        	return getObject().isDataAvailable();
        }
        
        /**
         *This returns the url of the file.
         *It's only available after you Save or after you get the file from a ParseObject.
         */
        public String getURL()
        {
        	return getObject().getUrl();
        }
        
        /**
         *Saves the object on the server.
         *The DoneSave event will be raised when saving completes.
         *If you want to ignore the event then pass an empty String for EventName.
         *If an error occurs Success will be False and a ParseException with a message describing the error will be placed in LastException.
         *A ParseFile must be saved before it can be added to a ParseObject
         */
        public void Save(final BA ba, final String EventName, final int TaskID) {
            if (EventName.length() == 0) {
                getObject().saveInBackground();
            }
            else {
                getObject().saveInBackground(new SaveCallback() {
        
                    @Override
                    public void done(ParseException e) {
                        setLastException(ba, e);
                        ba.raiseEventFromDifferentThread(getObject(), null, 0, EventName.toLowerCase(BA.cul) + "_donesave", 
                            true,
                            new Object[] {e == null, TaskID});
                    }
                    
                });
            }
        }
      
        /**
         *Gets the data for the object from the server.
         *The DoneGet event will be raised when saving completes.
         *If an error occurs Success will be False and a ParseException with a message describing the error will be placed in LastException.
         */
        public void Get(final BA ba, final String EventName, final int TaskID) {
        	getObject().getDataInBackground(new GetDataCallback() {        
	            @Override
	            public void done(byte[] data, ParseException e) {
	                setLastException(ba, e);
	                ba.raiseEventFromDifferentThread(getObject(), null, 0, EventName.toLowerCase(BA.cul) + "_doneget", 
	                    true,
	                    new Object[] {e == null, data, TaskID});
	            }            
        	});
        }
    }
        
        
	//###########################################################################
	// ParseGeoPoint
	//###########################################################################

	/**
	 *ParseGeoPoint represents a latitude/longitude point that may be associated with a key in a ParseObject
	 *or used as a reference point for geo queries. This allows proximity based queries on the key.
	 *
	 *Only one Key in a class may contain a GeoPoint.
	 *
	 *Points should not equal or exceed the extreme ends of the allowable ranges.
	 *Latitude should not be -90.0 or 90.0.
	 *Longitude should not be -180.0 or 180.0.
	 *Attempting to set latitude or longitude out of bounds will cause an error.
	 */
	@ShortName("ParseGeoPoint")
	public static class ParseGeoPointWrapper extends AbsObjectWrapper<ParseGeoPoint>
	{

		/**
		 *Returns the version of the library.
		 */
		public double getVersion()
		{
			return version;
		}

		/**
		 *Initializes the object and sets latitude and longitude.
		 *Latitude should not be -90.0 or 90.0.
		 *Longitude should not be -180.0 or 180.0.
		 */
		public void Initialize(double latitude, double longitude)
		{
			setObject(new ParseGeoPoint(latitude, longitude));
		}

		/**
		 * Initializes the object and sets both latitude and longitude 0.
		 */
		public void Initialize2()
		{
			setObject(new ParseGeoPoint());
		}

		/**
		 *Gets or sets the latitude for this GeoPoint.
		 *Latitude should not be -90.0 or 90.0.
		 */
		public double getLatitude()
		{
			return getObject().getLatitude();
		}
		public void setLatitude(double latitude)
		{
			getObject().setLatitude(latitude);
		}
				
		/**
		 *Gets or sets the longitude for this GeoPoint.
		 *Longitude should not be -180.0 or 180.0.
		 */
		public double getLongitude()
		{
			return getObject().getLongitude();
		}
		public void setLongitude(double longitude)
		{
			getObject().setLongitude(longitude);
		}

		/**
		 *Get the distance between this point and another geopoint in kilometers.
		 */
		public double DistanceInKilometersTo(ParseGeoPoint GeoPoint)
		{
			return getObject().distanceInKilometersTo(GeoPoint);
		}

		/**
		 *Get the distance between this point and another geopoint in miles.
		 */
		public double DistanceInMilesTo(ParseGeoPoint GeoPoint)
		{
			return getObject().distanceInMilesTo(GeoPoint);
		}

		/**
		 *Get distance in radians between this point and another GeoPoint.
		 *This is the smallest angular distance between the two points.
		 */
		public double DistanceInRadiansTo(ParseGeoPoint GeoPoint)
		{
			return getObject().distanceInRadiansTo(GeoPoint);
		}
		
	}


	//###########################################################################
	// ParseRelation
	//###########################################################################
	
	/**
	 *A class that is used to access all of the children of a many-to-many relationship.
	 *Each instance of Parse.Relation is associated with a particular parent object and key.
	 *A ParseRelation is obtained by invoking GetRelation on a ParseObject.
	 */
	@ShortName("ParseRelation")
	public static class ParseRelationWrapper extends AbsObjectWrapper<ParseRelation>
	{

		/**
		 *Returns the version of the library.
		 */
		public double getVersion()
		{
			return version;
		}


		/**
		 *Adds an object to this relation.
		 */
		public void Add(ParseObject Object)
		{			
			getObject().add(Object);
		}
		
		/**
		 *Removes an object from this relation.
		 */
		public void Remove(ParseObject Object)
		{			
			getObject().remove(Object);
		}
		
		/**
		 *Gets a ParseQuery that can be used to query the objects in this relation and that
		 *restricts the results to objects in this relation.
		 */
		public ParseQueryWrapper GetQuery()
		{
			ParseQueryWrapper pqw = new ParseQueryWrapper();
			pqw.setObject(getObject().getQuery());
			return pqw;
		}	
        
    } 
	
	
	//###########################################################################
	// ParsePush
	//###########################################################################
	
	/**
	 *A class that is used to access all of the children of a many-to-many relationship.
	 *Each instance of Parse.Relation is associated with a particular parent object and key.
	 *A ParseRelation is obtained by invoking GetRelation on a ParseObject.
	 */
    @Events(values={"DoneSend(Success As Boolean, TaskID As Int)" })
    @ShortName("ParsePush")
	public static class ParsePushWrapper extends AbsObjectWrapper<ParsePush>
	{

		/**
		 *Returns the version of the library.
		 */
		public double getVersion()
		{
			return version;
		}

		/**
		 *Initializes the ParsePush object.
		 *The default channel is the empty string, also known as the global broadcast channel,
		 *but this value can be overridden using SetChannel(String) 
		 */
		public void Initialize()
		{
			setObject(new ParsePush());
		}		
		
		/**
		 *Clears both expiration values, indicating that the notification should never expire.  
		 */
		public void ClearExpiration()
		{
			getObject().clearExpiration();
		}		
		
		/**
		 *Sets the channel on which this push notification will be sent. 
		 */
		public void SetChannel(String Channel) 
		{
			getObject().setChannel(Channel);
		}
        		
		/**
		 *Sets the timestamp at which this notification should expire, in seconds (UTC).
		 *This notification will be sent to devices which are either online at the time the notification is sent,
		 *or which come online before the expiration time is reached.
		 *Because device clocks are not guaranteed to be accurate, most applications should instead use ParsePush.setExpirationTimeInterval.
		 *Note that the underlying ParsePush method wants a Unix time so this method takes an Android time in ticks and converts it.
		 */
		public void SetExpirationTime(long TimeTicks) 
		{
			getObject().setExpirationTime(TimeTicks/DateTime.TicksPerSecond);
		}
         	
		/**
		 *Sets the time interval after which this notification should expire, in seconds.
		 *This notification will be sent to devices which are either online at the time the notification is sent,
		 *or which come online within the given number of seconds of the notification being received by Parse's server.
		 *An interval which is less than or equal to zero indicates that the message should only be sent to devices which are currently online. 
		 */
		public void SetExpirationTimeInterval(long Interval) 
		{
			getObject().setExpirationTimeInterval(Interval);
		}
		
		/**
		 *Sets the message that will be shown in the notification. 
		 */
		public void SetMessage(String Message)
		{
			getObject().setMessage(Message);
		}
		
        /**
         *Send the push notification. The DoneSend event will be raised when sending completes.
         *If an error occurs Success will be False and a ParseException with a message describing the error will be placed in LastException.
         *If you want to ignore the event then pass an empty String for EventName.
         */
        public void Send(final BA ba, final String EventName, final int TaskID) {
            if (EventName.length() == 0) {
                getObject().sendInBackground();
            }
            else {
                getObject().sendInBackground(new SendCallback() {        
                    @Override
                    public void done(ParseException e) {
                        setLastException(ba, e);
                        ba.raiseEventFromDifferentThread(getObject(), null, 0, EventName.toLowerCase(BA.cul) + "_donesend", 
                            true,
                            new Object[] {e == null, TaskID});
                    }
                    
                });
            }
        }
        
    }
        
    
}